theory PCL = TermList:

ML "set show_types"


(* Enc and Sig need to know about termlists, so they go here instead of PCLTerm.thy *)
consts
	Enc				::	"[key,termlist] => term"	("ENC{_}'(_')")
	Sig				::	"[key,termlist] => term"	("SIG{_}'(_')")

	NameKey			::	"name \<Rightarrow> key"				("key'(_')")

syntax
(* Now these are a datatype
(* Thread's can be variable or name/sid pairs *)
	ThreadPair		::	"[name,sid] => thread"					("<_,_>")
	Thread			::	"thread => thread"						("_")
*)
(* Action Formulas *)
	Send			::	"[thread,termlist] => ActionFormula"	("Send'(_,_')")
	Receive			::	"[thread,termlist] => ActionFormula"	("Receive'(_,_')")
	New				::	"[thread,termlist] => ActionFormula"	("New'(_,_')")
	Decrypt			::	"[thread,termlist] => ActionFormula"	("Decrypt'(_,_')")
	Verify			::	"[thread,termlist] => ActionFormula"	("Verify'(_,_')")
consts
	Send			::	"[thread,termlist] => bool"
	Receive			::	"[thread,termlist] => bool"
	New				::	"[thread,termlist] => bool"
	Decrypt			::	"[thread,termlist] => bool"
	Verify			::	"[thread,termlist] => bool"
(* Formulas *)
syntax
	""				::	"ActionFormula => bool"				("_" 50) 
	Previous		::	"[ActionFormula,ActionFormula] => bool"	(infixr "<<" 50)
consts
	Previous		::	"[bool,bool] => bool"
	Has				::	"[thread,termlist] => bool"			("Has'(_,_')")
	Fresh			::	"[thread,termlist] => bool"			("Fresh'(_,_')")
	FirstSend		::	"[thread,termlist,termlist] => bool"	("FirstSend'(_,_,_')")
	Honest			::	"[name] => bool"			("Honest'(_')")
	Start			::	"[thread] => bool"					("Start'(_')")
	(* Contains Predicate is in TermList.thy *)

(* actions *)
(*
	noopA			::	"action"						("noop")
	sendA			::	"termlist => action"			("send _")
	receieveA		::	"termlist => action"			("receive _")
	newA			::	"termlist => action"			("new _")
	matchA			::	"[termlist,'a::PCLTerm] => action"	("match _ '/ _")
	
	StrandList		::	"[strand,strand] => strand"		("_ ; _ " [60,40] 41)
	StrandUnit		::	"action => strand"				("_")
*)

datatype action = 
	  Noop							("noop")
	| SendA termlist				("send _")
	| ReceiveA termlist				("receive _")
 	| newA	termlist				("new _")
	| matchA termlist "term"	("match _ '/ _")
	

types strand = "action list"


(* Modal Formulas *)
consts
	ModalFormula	::	"[bool,strand,thread,bool] => prop"				("MF'(_,_,_,_')")




(*
consts
	Cord			::	"[strand,thread] => cord"			("_ '_ _")
*)


axioms

(* Conventions
 * -----------
 *
 * m,t,u,v,w,x	terms
 * s			session id
 * P,Q,R		action formulas and formulas
 * S			strands
 * C,D,E		cords 
 * X,Y,Z		threads
 *)


	AA1S:	"MF( P, [send t], X, Send(X,t))"
	AA1R:	"MF( P, [receive t], X, Receive(X,t))"
	AA1N:	"MF( P, [new t], X, New(X,t))"
	AA1D:	"MF( P, [match u / ENC{k}(t)], X, Decrypt(X,(ENC{k}(t) ## [])) )"
	AA1V:	"MF( P, [match u / SIG{k}(t)], X, Verify(X,(SIG{k}(t) ## [])) )"

	AA2S:	"MF( Start(X), [noop], X, ~Send(X,t))"
	AA2R:	"MF( Start(X), [noop], X, ~Receive(X,t))"
	AA2N:	"MF( Start(X), [noop], X, ~New(X,t))"
	AA2D:	"MF( Start(X), [noop], X, ~Decrypt(X,t))"
	AA2V:	"MF( Start(X), [noop], X, ~Verify(X,t))"
	
	(*	AA3 has two cases:  the action is a send or it's not.  if it's a send
	 *	we need to check the truth of the Contains predicate.  Otherwise (if
	 *	the action is not a send), we automatically know ~Send(X,t)
	 *)


	AA3S:	"\<lbrakk> Contains(u,t) \<rbrakk> \<Longrightarrow> MF( ~Send(X,t), [send u], X, ~Send(X,t))"

	AA3R:	"MF( ~Send(X,t), [receive u], X, ~Send(X,t))"
	AA3N:	"MF( ~Send(X,t), [new u], X, ~Send(X,t))"
	AA3D:	"MF( ~Send(X,t), [match  u/ENC{k}(v)], X, ~Send(X,t))"
	AA3V:	"MF( ~Send(X,t), [match u/SIG{k}(v)], X, ~Send(X,t))"
	AA3No:	"MF( ~Send(X,t), [No], X, ~Send(X,t))"


	AA5S:	"\<lbrakk> ~ Contains(u,v) \<rbrakk> \<Longrightarrow> MF( ~(Send(X,t) & Contains(u,t)), [send v], X, ~(Send(X,t) & Contains(u,t)))"
	AA5R:	"MF( ~(Send(X,t) & Contains(u,t)), [receive v], X, 	~(Send(X,t) & Contains(u,t)))"
	AA5N:	"MF( ~(Send(X,t) & Contains(u,t)), [new v], X, 	~(Send(X,t) & Contains(u,t)))"	
	AA5D:	"MF( ~(Send(X,t) & Contains(u,t)), [match v/ENC{k}(w)], X,	~(Send(X,t) & Contains(u,t)))"	
	AA5V:	"MF( ~(Send(X,t) & Contains(u,t)), [match v/SIG{k}(w)], X, 	~(Send(X,t) & Contains(u,t)))"
	AA5No:	"MF( ~(Send(X,t) & Contains(u,t)), [noop], X, 	~(Send(X,t) & Contains(u,t)))"

	AN2:	"MF( P, [new t], X, Has(Y,t) \<longrightarrow> (Y = X))"
	AN3:	"MF( P, [new t], X, Fresh(X,t))"
	
	ORIG:	"New(X,t) \<longrightarrow> Has(X,t)"
	REC:	"Receive(X,t) \<longrightarrow> Has(X,t)"
	TUP:	"Has(X,t) & Has(X,u) \<longrightarrow> Has(X,(t @ u))"
	ENC:	"Has(X,t) & Has(X,((k::key) ## [])) \<longrightarrow> Has(X,(ENC{k}(t) ## []))"
	PROJ:	"Has(X,(t @ u)) \<longrightarrow> Has(X,t) & Has(X,u)"
	DEC:	"Has(X,(ENC{k}(t) ## [])) & Has(X,(k ## [])) \<longrightarrow> Has(X,t)"

	SEC:	"Honest(X) & (EX s. Decrypt(<Y,s>,(ENC{key(X)}(t) ## []))) \<longrightarrow> (Y = X)"
	VER:	"Honest(X) & Verify(<Y,s1>,(SIG{key(X)}(t)) ## []) & (Y ~= X) \<longrightarrow>  (EX m. (EX s2. Send(<X,s2>,m) & Contains(m,(SIG{key(X)}(t) ## []))))"
	
	G1:		"\<lbrakk> MF( P, C, X, Q) ; MF( P, C, X, R) \<rbrakk> \<Longrightarrow> MF( P, C, X, Q & R)"
	G2:		"\<lbrakk> MF( P, C, X, R) ; MF( Q, C, X, R) \<rbrakk> \<Longrightarrow> MF( P & Q, C, X, R)"
	G3:		"\<lbrakk> MF( Pa, C, X, Qa) ; (Pb \<longrightarrow> Pa) ; (Qa \<longrightarrow> Qb) \<rbrakk> \<Longrightarrow> MF( Pb, C, X, Qb)"
	G4:		"\<lbrakk> Q \<rbrakk> \<Longrightarrow> MF( P, C, X, Q)"

	SEQ:	"\<lbrakk> MF( P, [Sa], X, Q) ;  MF( Q, Sb, X, R) \<rbrakk> \<Longrightarrow> MF( P, Sa # Sb, X, R)"

(*	use "tactics.ML" *)

	P1S:	"MF( Send(X,t), S, X, Send(X,t))"
	P1R:	"MF( Receive(X,t), S, X, Receive(X,t))"
	P1N:	"MF( New(X,t), S, X, New(X,t))"
	P1D:	"MF( Decrypt(X,t), S, X, Decrypt(X,t))"
	P1V:	"MF( Verify(X,t), S, X, Verify(X,t))"
	P1H:	"MF( Has(X,t), S, X, Has(X,t))"
	P1FS:	"MF( FirstSend(X,ta,tb), S, X, FirstSend(X,ta,tb))"

(* brute force.  has to be a better way *)

	P1SS:	"MF((Send(X,ta) << Send(X,tb)), S, X, (Send(X,ta) << Send(X,tb)))"
	P1SR:	"MF((Send(X,ta) << Receive(X,tb)), S, X, (Send(X,ta) << Receive(X,tb)))"
	P1SN:	"MF((Send(X,ta) << New(X,tb)), S, X, (Send(X,ta) << New(X,tb)))"
	P1SD:	"MF((Send(X,ta) << Decrypt(X,tb)), S, X, (Send(X,ta) << Decrypt(X,tb)))"
	P1SV:	"MF((Send(X,ta) << Verify(X,tb)), S, X, (Send(X,ta) << Verify(X,tb)))"

	P1RS:	"MF((Receive(X,ta) << Send(X,tb)), S, X, (Receive(X,ta) << Send(X,tb)))"
	P1RR:	"MF((Receive(X,ta) << Receive(X,tb)), S, X, (Receive(X,ta) << Receive(X,tb)))"
	P1RN:	"MF((Receive(X,ta) << New(X,tb)), S, X, (Receive(X,ta) << New(X,tb)))"
	P1RD:	"MF((Receive(X,ta) << Decrypt(X,tb)), S, X, (Receive(X,ta) << Decrypt(X,tb)))"
	P1RV:	"MF((Receive(X,ta) << Verify(X,tb)), S, X, (Receive(X,ta) << Verify(X,tb)))"

	P1NS:	"MF((New(X,ta) << Send(X,tb)), S, X, (New(X,ta) << Send(X,tb)))"
	P1NR:	"MF((New(X,ta) << Receive(X,tb)), S, X, (New(X,ta) << Receive(X,tb)))"
	P1NN:	"MF((New(X,ta) << New(X,tb)), S, X, (New(X,ta) << New(X,tb)))"
	P1ND:	"MF((New(X,ta) << Decrypt(X,tb)), S, X, (New(X,ta) << Decrypt(X,tb)))"
	P1NV:	"MF((New(X,ta) << Verify(X,tb)), S, X, (New(X,ta) << Verify(X,tb)))"

	P1DS:	"MF((Decrypt(X,ta) << Send(X,tb)), S, X, (Decrypt(X,ta) << Send(X,tb)))"
	P1DR:	"MF((Decrypt(X,ta) << Receive(X,tb)), S, X, (Decrypt(X,ta) << Receive(X,tb)))"
	P1DN:	"MF((Decrypt(X,ta) << New(X,tb)), S, X, (Decrypt(X,ta) << New(X,tb)))"
	P1DD:	"MF((Decrypt(X,ta) << Decrypt(X,tb)), S, X, (Decrypt(X,ta) << Decrypt(X,tb)))"
	P1DV:	"MF((Decrypt(X,ta) << Verify(X,tb)), S, X, (Decrypt(X,ta) << Verify(X,tb)))"

	P1VS:	"MF((Verify(X,ta) << Send(X,tb)), S, X, (Verify(X,ta) << Send(X,tb)))"
	P1VR:	"MF((Verify(X,ta) << Receive(X,tb)), S, X, (Verify(X,ta) << Receive(X,tb)))"
	P1VN:	"MF((Verify(X,ta) << New(X,tb)), S, X, (Verify(X,ta) << New(X,tb)))"
	P1VD:	"MF((Verify(X,ta) << Decrypt(X,tb)), S, X, (Verify(X,ta) << Decrypt(X,tb)))"
	P1VV:	"MF((Verify(X,ta) << Verify(X,tb)), S, X, (Verify(X,ta) << Verify(X,tb)))"

(*
consts

	AFPred	:: "ActionFormula => bool";

axioms

	P1Test:	"[| AFPred(a) ; AFPred(b) |] ==> A & B";
*)

	FS1:	"[| Contains(ta,tb) |] ==> MF( Fresh(X,ta), [send tb], X, FirstSend(X,ta,tb))"

	FS2S:	"(FirstSend(X,ta,tb) & Send(Y,tc)) --> Send(X,tb) << Send(Y,tc)"
	FS2R:	"(FirstSend(X,ta,tb) & Receive(Y,tc)) --> Send(X,tb) << Receive(Y,tc)"
	FS2N:	"(FirstSend(X,ta,tb) & New(Y,tc)) --> Send(X,tb) << New(Y,tc)"
	FS2D:	"(FirstSend(X,ta,tb) & Decrypt(Y,tc)) --> Send(X,tb) << Decrypt(Y,tc)"
	FS2V:	"(FirstSend(X,ta,tb) & Verify(Y,tc)) --> Send(X,tb) << Verify(Y,tc)"

	
end


