(*	Title:	pcl/NewCord.thy
	Author:	Dan Auerbach and Cary Kempston
	ID:	$Id: NewCord.thy,v 1.18 2004/08/26 09:14:58 auerbach Exp $
*)
header {* Protocol Composition Logic *}

theory NewCord = FOL:

global


subsection {* Types for things in PCL *}

typedecl	name
typedecl	sid
typedecl	key
typedecl	CTerm	
typedecl	thread
typedecl 	nonce
typedecl	strand
typedecl	cord


nonterminals

	CTermlist
	CTermI
	threadI
	actformI
	strandI
	cordI
	keyI


text {* Needed for equality in axioms *}

arities

	thread	:: "term"
	CTerm	:: "term"
	key		:: "term"
	sid		:: "term"
	name	:: "term"
	strand	:: "term"
	cord	:: "term"

subsection {* The Grammer of PCL *}


subsubsection {* Action formulas *}

consts
	PSend	::	"[thread, CTerm] => o"
	PReceive::	"[thread, CTerm] => o"
	PNew	::      "[thread, CTerm] => o"
	PDecrypt::      "[thread, CTerm] => o"
	PVerify	::      "[thread, CTerm] => o"

subsubsection {* Formulas *}	

consts
	PHas		:: "[thread, CTerm] => o"
	PFirstSend	:: "[thread, CTerm] => o"
	PPrevious	:: "[o, o] => o"
   	PFresh		:: "[thread, CTerm] => o"         
	PHonest		:: "name  => o"                                     
   	PContains	:: "[CTerm, CTerm] => o"                   
   	PStart		:: "thread => o"    

subsubsection {* Modal formulas *}

consts
	ModalFormula	:: "[o, cord, o]=>prop"                       

subsubsection {* terms *}

consts
	MThread	::	"[name, sid] => thread"
	TEnc	::	"[key, CTerm] => CTerm"
	TSig    ::	"[key, CTerm] => CTerm"
	NameKey	::	"name => key"			(* using a name in ENC or SIG *)
 
	(* nonces, etc, as terms *)
	TNonce	::	"nonce =>CTerm"
	TName	::	"name => CTerm"
	TThread	::	"thread => CTerm"
	TKey	::	"key => CTerm"

	(* term lists *)
	CTermList	::	"[CTerm, CTerm] => CTerm"
	CTermParen	::	"CTerm => CTerm"


subsubsection {* Cord Calculus *}

consts
	Noop            :: strand                       
   	ActionSend      :: "CTerm => strand"              
  	ActionReceive   ::      "CTerm => strand"
 	ActionNew       ::      "CTerm => strand"          
   	ActionMatch     ::      "[CTerm, CTerm] => strand"         
	StrandList      ::  "[strand, strand] => strand"

	StrandSplit		::	"[strand, strand] => strand" (* needed by axiom SEQ *)

	Cord            :: "[thread, strand] => cord"     

        (*special productions for AA4*)
	PInorder :: "[thread, strand] => o"                             
	PrevActStrand   :: "[o, thread, strand] => o"  

subsection {* Parsing syntax *}

subsubsection {* Action formulas *}

syntax
	""	 ::	"id => actformI" ("_" 0)
	PSend   ::	"[threadI, CTermlist] => actformI" ("Send'(_,_')")
	PReceive::	"[threadI, CTermlist] => actformI" ("Receive'(_,_')" )
	PNew    ::    "[threadI, CTermlist] => actformI" ("New'(_,_')" )
	PDecrypt::    "[threadI, CTermlist] => actformI" ("Decrypt'(_,_')"  )
	PVerify ::    "[threadI, CTermlist] => actformI" ("Verify'(_,_')" )


subsubsection {* Formulas *}

syntax
	"PHas"            :: "[threadI, CTermlist] => o" ("Has'(_,_')")
	"PFirstSend"      :: "[threadI, CTermlist] => o" ("FirstSend'(_,_')")
	"PPrevious"        :: "[actformI, actformI] => o" ("_ < _")
	"PFresh"           :: "[threadI, CTermlist] => o" ("Fresh'(_,_')")
	"PHonest"          :: "idt  => o" ("Honest'(_')")
	"PContains"        :: "[CTermlist, CTermlist] => o" ("Contains'(_,_')")
	"PStart"           :: "threadI => o" ("Start'(_')")

	""	::	"actformI => o" ("_")

subsubsection {* Modal formulas *}

syntax	
	"ModalFormula"	:: "[o, cordI, o] => prop"	("{_,_,_}" [5, 0, 5] 90)

subsubsection {* terms *}
	
syntax
	"MThread"	::	"[idt, idt] => threadI"	("<_,_>" 80) 
   	"TEnc"		::  "[keyI, CTermlist] => CTermI"	  ("ENC[_]'(_')")
	"TSig"		::	"[keyI, CTermlist] => CTermI"   ("SIG[_]'(_')")
	"NameKey"	::	"idt => keyI"				("key'(_')")
 
	"TNonce"	::	"idt => CTermI"	("nonce'(_')")
	"TName"		::	"idt => CTermI"	("na'(_')")
	"TThread"	::	"idt => CTermI"	("thread'(_')")
	"TKey"		::	"idt => CTermI"	("key'(_')")


	"CTermList"	::	"[CTermlist, CTermlist] => CTermlist"	("_ , _" [60, 40] 40)
	"CTermParen"	::	"CTermlist => CTermlist"		("'( _ ')")
	""			:: "CTermI => CTermlist" ("_" 60) 			

	""		:: 	"idt => keyI"		("_")
	""		:: "idt => CTermI"		("_" 80)
	""		:: "idt => threadI"		("_" 60)

	(* cord calculus *)

	""					:: "idt => strandI" ("_")
	"Noop"         	   	:: strandI                               ("noop" 60)
	"ActionSend"   	   	:: "CTermlist => strandI"               ("send _" 60)
	"ActionReceive"   	::      "CTermlist => strandI"  ("receive _" 60)
	"ActionNew"     	::      "CTermlist => strandI"          ("new _" 60)
	"ActionMatch"   	::      "[CTermlist, CTermI] => strandI"   ("match _ '/ _" 60)
    
	"StrandList"      	::  "[strandI, strandI] => strandI"          ("_;_" [60, 40] 40)

	"StrandSplit"	 	:: "[strandI, strandI] => strandI"	("_ * _" 0)

	""		  			:: "idt => cordI" ("_")
	"Cord"		  		:: "[threadI, strandI] => cordI"	("_[_]" [40, 0] 0)

        (*special productions for AA4*)

	"PInorder" :: "[threadI, strandI] => o" ("Inorder'(_: _')")
	"PrevActStrand"  :: "[o, threadI, strandI] => o" (" _ <# _:_" [5, 0, 0] 70)




(* Conventions
 * -----------
 *
 * t,u,v,x		CTerms
 * s			session id
 * P,Q,R		action formulas
 * S			Strands
 * C,D,E		cords
 * X,Y,Z		threads
 *)

subsection {* Proof System *}

subsubsection {* Axioms for Protocol Actions *}

axioms
	AA1S:	"{P, X[send t], Send(X,t)}"
	AA1N:	"{P, X[new t], New(X,t)}"
	AA1R:	"{P, X[receive t], Receive(X,t)}"
	AA1D:	"{P, X[match u / ENC[k](t) ], Decrypt(X,ENC[k](t))}"
	AA1V:	"{P, X[match u / SIG[k](t) ], Verify(X,SIG[k](t))}"

	AA2S:	"{P, X[noop], ~Send(X,t)}"
	AA2N:	"{P, X[noop], ~New(X,t)}"
	AA2R:	"{P, X[noop], ~Receive(X,t)}"
	AA2D:	"{P, X[noop], ~Decrypt(X,t)}"
	AA2V:	"{P, X[noop], ~Verify(X,t)}"

	AA3S:   "(~Contains(u , t)) ==> {~(Send(X,v) & Contains(v, t)), X[send u], ~Send(X,v)}"
	AA3N:    "{~Send(X,t), X[new u], ~Send(X,t)}"
	AA3R:    "{~Send(X,t), X[receive  u], ~Send(X,t)}"
	AA3D:    "{~Send(X,t), X[match u/ENC[k](t)], ~Send(X,t)}"
	AA3V:    "{~Send(X,t), X[match u/SIG[k](t)], ~Send(X,t)}"

	(*Unwieldy Implementation of AA4 *)

        (*Order axioms for a predicate coming before a strand*)
	ORDS:     "P <# X: send t; S --> P < Send(X,t)"
	ORDR:     "P <# X: receive t; S --> P < Receive(X,t)" 
	ORDN:     "P <# X: new t; S --> P < New(X,t)" 
	ORDD:     "P <# X: match t/SIG[k](u); S --> P < Decrypt(X,t)" 
	ORDV:     "P <# X: match t/ENC[k](u); S --> P < Verify(X,t)"
	ORDPUSH:  "(P <# X: a; S) --> (P <# X: S)"   
        
       (*the basic AA4 axiom*)
	AA4:     "{P, X[S], Inorder(X: S)}"

        (*axioms relating order to inorder*)
	AA4S:    "Inorder(X: send t; S) --> Inorder(X:S) & (Send(X,t) <# X: S)"
	AA4R:    "Inorder(X: receive t; S) --> Inorder(X: S) & (Receive(X,t) <# X: S)"
	AA4N:    "Inorder(X: new t; S) --> Inorder(X: S) & (New(X,t) <# X: S)"
	AA4D:    "Inorder(X: match t/SIG[k](u); S) --> Inorder(X: S) & (Decrypt(X,t) <# X: S)"
	AA4V:    "Inorder(X: match t/ENC[k](u); S) --> Inorder(X: S) & (Verify(X,t) <# X: S)" 

        (*for AA4: can add/remove noops at the end of a strand*)
    ADDNOOPR:  "{P, X[S; noop], Q} ==> {P, X[S], Q}"
    REMNOOPR:  "{P, X[S], Q} ==> {P, X[S;noop], Q}"
    ADDNOOPL:  "{P, X[noop; S], Q} ==> {P, X[S], Q}"
    REMNOOPL:  "{P, X[S], Q} ==> {P, X[noop; S], Q}"


	AN2:	"{P, X[new t], Has(Y,x) --> X = Y}" 	
	AN3:	"{P, X[new t], Fresh(X,t)}"

	(* still need ARP *)

subsubsection {* Possession Axioims *}

axioms
	ORIG:	"New(X,t) --> Has(X,t)"
	REC:	"Receive(X,t) --> Has(X,t)"
	TUP:	"Has(X,t) & Has(X,u) --> Has(X,t,u)"
	ENC:	"Has(X,t) & Has(X,key(k)) --> Has(X,ENC[k](t))" 
	PROJ:	"Has(X,(t,u)) --> Has(X,t) & Has(X,u)"
	DEC:	"Has(X,ENC[k](t)) & Has(X,key(k)) --> Has(X,t)" 

subsubsection {* Encryption and Signature *}

axioms
	SEC:    "Honest(X) & Decrypt(<Y,s>, ENC[key(X)](t)) --> Y = X" 
	VER:    "(Honest(X) & Verify(<Y, s1>, SIG[key(X)](t)) & X ~= Y) --> ((Send(<X,s2>, u) & Contains(u, SIG[key(X)](t))))"    

subsubsection {* Generic Rules *}

axioms
	G1:	"[| {P, C, Q} ; {P, C, R} |] ==> {P, C, Q & R}"
        G2:     "[| {P, C, R}; {Q, C, R} |] ==> {P & Q, C, R}"
	G3:	"[| {P, C, Q} ; P-->P1; Q-->Q1 |] ==> {P1, C, Q1}"
	G4:	"Q::o ==> {P, C, Q}"

subsubsection {* Sequencing Rule *}

axioms
	OLDSEQ:	"[|{P, X[S1], Q} ; {Q, X[S2], R} |] ==> {P, X[S1;  S2], R}"

	SEQ:	"[|{P, X[S1], Q} ; {Q, X[S2], R} |] ==> {P, X[S1 * S2], R}"

	MSR:	"[|{P, X[S1; S2 * S3], Q} |] ==> {P, X[S1 * S2; S3], Q}" 
	MSL:	"[|{P, X[S1 * S2; S3], Q} |] ==> {P, X[S1 ; S2 * S3], Q}"

	InsertStar:	"[| {P, X[S1 *  S2], Q} |] ==> {P, X[S1 ; S2], Q}"
	RemoveStar:	"[| {P, X[S1 ; S2], Q} |] ==> {P, X[S1 * S2], Q}" 

subsubsection {* Preservation Axioms *}

axioms
	P1S:	"{Send(X,t), X[S], Send(X,t)}"
	P1R:	"{Receive(X,t), X[S], Receive(X,t)}"
	P1N:	"{New(X,t), X[S], New(X,t)}"
	P1D:	"{Decrypt(X,t), X[S], Decrypt(X,t)}"
	P1V:	"{Verify(X,t), X[S], Verify(X,t)}"
	P1H:	"{Has(X,t), X[S], Has(X,t)}"
	P1F:	"{FirstSend(X,t), X[S], FirstSend(X,t)}"
        P1P:    "{a < b, X[S], a < b}"

	P2S:	"~(Contains(u,t)) ==> {Fresh(X,t), X[send u], Fresh(X,t)}"
	P2R:	"{Fresh(X,t), X[receive u], Fresh(X,t)}"
	P2N:	"{Fresh(X,t), X[new u], Fresh(X,t)}"
 	P2M:	"{Fresh(X,t), X[match  u/v], Fresh(X,t)}"

	(* P3 not implemented yet - needed for Diffe Hellmen *)

subsubsection {* Axioms and Rules for Termporal Ordering *}

axioms
	FS1:	"Contains(t2, t1) ==> {Fresh(X,t1), X[send t2], FirstSend(X,t1,t2)}"

	FS2S:	"Contains(t3, t1) ==> FirstSend(X,t1,t2) & Send(Y,t3) &  X ~= Y --> Send(X,t2) < Send(Y,t3)"
	FS2R:	"Contains(t3, t1)==> FirstSend(X,t1,t2) & Receive(Y,t3)--> Send(X,t2) < Receive(Y,t3)" (*needs to be updated*)
	FS2N:	"Contains(t3, t1)==> FirstSend(X,t1,t2) & New(Y,t3)  & X ~= Y --> Send(X,t2) < New(Y,t3)"
	FS2D:	"Contains(t3, t1)==> FirstSend(X,t1,t2) & Decrypt(Y,t3) & X ~= Y --> Send(X,t2) < Decrypt(Y,t3)"
	FS2V:	"Contains(t3, t1)==> FirstSend(X,t1,t2) & Verify(Y,t3) & X ~= Y --> Send(X,t2) < Verify(Y,t3)"

        ACT:    "a < b --> a & b"


subsubsection {*Honesty Rule for Challenge-Response*}

axioms

        HON: "[| {Start(<Y,s>), <Y,s>[noop], P}; {P, <Y,s>[new y; send na(Y), na(X), y], P}; {P, <Y,s>[receive (x, na(X),na(Y)),z; match z/SIG[key(X)](x,y,na(Y))], P}; {P, <Y,s>[receive (x, na(X), na(Y)); new y; send (y, na(Y), na(X), SIG[key(Y)](y,x,na(X)))], P} |] ==> Honest(Y) --> P"

end
